@@ -15,6 +15,7 @@
 
 import json
 import logging
+import platform
 import pprint
 import select
 import six.moves
@@ -34,13 +35,15 @@ UPDATE_ALLOWED_KEYS = ['description', 'submit-type',
                        'content-merge', 'change-id',
                        'project-state',
                        'max-object-size-limit']
+IS_LINUX = platform.system() == "Linux"
 
 
 class GerritConnection(object):
     log = logging.getLogger("gerrit.GerritConnection")
 
     def __init__(self, username=None, hostname=None, port=29418,
-                 keyfile=None, connection_attempts=-1, retry_delay=5):
+                 keyfile=None, keep_alive=0, connection_attempts=-1,
+                 retry_delay=5):
 
         assert retry_delay >= 0, "Retry delay must be >= 0"
         self.username = username
@@ -49,6 +52,7 @@ class GerritConnection(object):
         self.keyfile = keyfile
         self.connection_attempts = int(connection_attempts)
         self.retry_delay = float(retry_delay)
+        self.keep_alive = keep_alive
 
     def connect(self):
         """Attempts to connect and returns the connected client."""
@@ -85,6 +89,8 @@ class GerritConnection(object):
                                username=self.username,
                                port=self.port,
                                key_filename=self.keyfile)
+                if self.keep_alive:
+                    client.get_transport().set_keepalive(self.keep_alive)
                 return client
             except (IOError, paramiko.SSHException) as e:
                 self.log.exception("Exception connecting to %s:%s",
@@ -108,7 +114,7 @@ class GerritWatcher(threading.Thread):
 
     def __init__(
             self, gerrit, username=None, hostname=None, port=None,
-            keyfile=None, connection_attempts=-1, retry_delay=5):
+            keyfile=None, keep_alive=0, connection_attempts=-1, retry_delay=5):
         """Create a GerritWatcher.
 
         :param gerrit: A GerritConnection instance to pass events to.
@@ -122,6 +128,7 @@ class GerritWatcher(threading.Thread):
             hostname or gerrit.connection.hostname,
             port or gerrit.connection.port,
             keyfile or gerrit.connection.keyfile,
+            keep_alive or gerrit.connection.keep_alive,
             connection_attempts,
             retry_delay
         )
@@ -131,22 +138,35 @@ class GerritWatcher(threading.Thread):
 
     def _read(self, fd):
         line = fd.readline()
-        data = json.loads(line)
-        self.log.debug("Received data from Gerrit event stream: \n%s" %
-                       pprint.pformat(data))
-        self.gerrit.addEvent(data)
+        if line:
+            try:
+                data = json.loads(line)
+                self.log.debug("Received data from Gerrit event stream: \n%s" %
+                               pprint.pformat(data))
+                self.gerrit.addEvent(data)
+            except json.decoder.JSONDecodeError:
+                self.log.debug("Can not parse data from Gerrit event stream: \n%s" %
+                               pprint.pformat(line))
+        else:
+            # blank read indicates EOF and the connection closed
+            self.state = DISCONNECTED
 
     def _listen(self, stdout, stderr):
         poll = select.poll()
         poll.register(stdout.channel)
         while True:
+            if self.state == DISCONNECTED:
+                return
             ret = poll.poll()
             for (fd, event) in ret:
                 if fd == stdout.channel.fileno():
-                    if event == select.POLLIN:
+                    # event is a bitmask
+                    # On Darwin non-critical OOB messages can be received
+                    if (IS_LINUX and event == select.POLLIN) or \
+                            (not IS_LINUX and event & select.POLLIN):
                         self._read(stdout)
                     else:
-                        raise Exception("event on ssh connection")
+                        raise Exception("event %s on ssh connection" % event)
 
     def _consume(self, client):
         """Consumes events using the given client."""
@@ -194,8 +214,16 @@ class GerritWatcher(threading.Thread):
 class Gerrit(object):
     log = logging.getLogger("gerrit.Gerrit")
 
-    def __init__(self, hostname, username, port=29418, keyfile=None):
-        self.connection = GerritConnection(username, hostname, port, keyfile)
+    def __init__(
+            self, hostname, username, port=29418, keyfile=None,
+            keep_alive_interval=0,
+            connection_attempts=-1,
+            retry_delay=5):
+        self.connection = GerritConnection(
+            username, hostname, port, keyfile,
+            keep_alive=keep_alive_interval,
+            connection_attempts=connection_attempts,
+            retry_delay=retry_delay)
         self.client = None
         self.watcher_thread = None
         self.event_queue = None
@@ -319,9 +347,19 @@ class Gerrit(object):
             cmd = 'gerrit ls-groups -v'
         else:
             cmd = 'gerrit ls-groups'
+        version = None
+        try:
+            version = self.parseGerritVersion(self.getVersion())
+        except Exception:
+            # If no version then we know version is old and should use -q
+            pass
+        if not version or version < (2, 14):
+            query_flag = '-q'
+        else:
+            query_flag = '-g'
         # ensure group names with spaces are escaped and quoted
         group = "\"%s\"" % group.replace(' ', r'\ ')
-        out, err = self._ssh(' '.join([cmd, '-q', group]))
+        out, err = self._ssh(' '.join([cmd, query_flag, group]))
         return list(filter(None, out.split('\n')))
 
     def listPlugins(self):
